@@ -21,6 +21,10 @@
 
 #include <config.h>
 
+#include <string.h>
+#include <glib/gi18n-lib.h>
+#include <gdesktop-enums.h>
+
 #include "cc-background-panel.h"
 #include "bg-wallpapers-source.h"
 #include "bg-pictures-source.h"
@@ -30,16 +34,19 @@
 #include "bg-flickr-source.h"
 #endif
 
-#include "gnome-wp-xml.h"
-#include "gnome-wp-item.h"
+#include "cc-background-item.h"
+#include "cc-background-xml.h"
 
-#include <string.h>
-#include <glib/gi18n-lib.h>
+#define WP_PATH_ID "org.gnome.desktop.background"
+#define WP_URI_KEY "picture-uri"
+#define WP_OPTIONS_KEY "picture-options"
+#define WP_SHADING_KEY "color-shading-type"
+#define WP_PCOLOR_KEY "primary-color"
+#define WP_SCOLOR_KEY "secondary-color"
 
 enum {
   COL_SOURCE_NAME,
   COL_SOURCE_TYPE,
-  COL_SOURCE_READONLY,
   COL_SOURCE,
   NUM_COLS
 };
@@ -65,8 +72,7 @@ struct _CcBackgroundPanelPrivate
 
   GnomeDesktopThumbnailFactory *thumb_factory;
 
-  GnomeWPItem *current_background;
-  gboolean current_source_readonly;
+  CcBackgroundItem *current_background;
   gint current_source;
 
   GCancellable *copy_cancellable;
@@ -198,7 +204,7 @@ cc_background_panel_finalize (GObject *object)
 
   if (priv->current_background)
     {
-      gnome_wp_item_free (priv->current_background);
+      g_object_unref (priv->current_background);
       priv->current_background = NULL;
     }
 
@@ -224,55 +230,34 @@ cc_background_panel_class_finalize (CcBackgroundPanelClass *klass)
 }
 
 static void
-source_update_edit_box (CcBackgroundPanelPrivate *priv)
+source_update_edit_box (CcBackgroundPanelPrivate *priv,
+			gboolean                  initial)
 {
-  if (priv->current_source == SOURCE_COLORS)
-    {
-      gtk_widget_hide (WID ("style-combobox"));
-      gtk_widget_show (WID ("style-pcolor"));
+  CcBackgroundItemFlags flags;
 
-      if (priv->current_background &&
-	  priv->current_background->shade_type == G_DESKTOP_BACKGROUND_SHADING_SOLID)
-        gtk_widget_hide (WID ("style-scolor"));
-      else
-        gtk_widget_show (WID ("style-scolor"));
-    }
+  flags = cc_background_item_get_flags (priv->current_background);
+
+  if ((flags & CC_BACKGROUND_ITEM_HAS_SCOLOR &&
+       priv->current_source != SOURCE_COLORS) ||
+      cc_background_item_get_shading (priv->current_background) == G_DESKTOP_BACKGROUND_SHADING_SOLID)
+    gtk_widget_hide (WID ("style-scolor"));
   else
-    {
-      if (!priv->current_source_readonly)
-        {
-          gtk_widget_show (WID ("style-pcolor"));
-          gtk_widget_hide (WID ("style-scolor"));
-          gtk_widget_show (WID ("style-combobox"));
-	}
-      else
-        {
-          gtk_widget_hide (WID ("style-pcolor"));
-          gtk_widget_hide (WID ("style-scolor"));
-          gtk_widget_hide (WID ("style-combobox"));
-	}
-    }
-}
+    gtk_widget_show (WID ("style-scolor"));
 
-static void
-setup_edit_box (CcBackgroundPanelPrivate *priv)
-{
-  g_assert (priv->current_background);
-
-  if (g_str_equal (priv->current_background->filename, "(none)"))
-    {
-      gtk_widget_hide (WID ("style-combobox"));
-      gtk_widget_show (WID ("style-pcolor"));
-
-      if (priv->current_background->shade_type == G_DESKTOP_BACKGROUND_SHADING_SOLID)
-        gtk_widget_hide (WID ("style-scolor"));
-      else
-        gtk_widget_show (WID ("style-scolor"));
-    }
+  if (flags & CC_BACKGROUND_ITEM_HAS_PCOLOR &&
+      priv->current_source != SOURCE_COLORS)
+    gtk_widget_hide (WID ("style-pcolor"));
   else
-    {
-      /* FIXME other cases */
-    }
+    gtk_widget_show (WID ("style-pcolor"));
+
+  if (flags & CC_BACKGROUND_ITEM_HAS_PLACEMENT ||
+      cc_background_item_get_uri (priv->current_background) == NULL)
+    gtk_widget_hide (WID ("style-combobox"));
+  else
+    gtk_widget_show (WID ("style-combobox"));
+
+  /* FIXME What to do if the background has a gradient shading
+   * and provides the colours? */
 }
 
 static void
@@ -298,39 +283,6 @@ source_changed_cb (GtkComboBox              *combo,
                            GTK_TREE_MODEL (bg_source_get_liststore (source)));
 }
 
-static void
-copy_finished_cb (GObject      *source_object,
-                  GAsyncResult *result,
-                  gpointer      pointer)
-{
-  GError *err = NULL;
-  CcBackgroundPanel *panel = (CcBackgroundPanel *) pointer;
-  CcBackgroundPanelPrivate *priv = panel->priv;
-
-  if (!g_file_copy_finish (G_FILE (source_object), result, &err))
-    {
-      if (err->code != G_IO_ERROR_CANCELLED)
-        g_warning ("Failed to copy image to cache location: %s", err->message);
-
-      g_error_free (err);
-    }
-
-  /* the panel may have been destroyed before the callback is run, so be sure
-   * to check the widgets are not NULL */
-
-  if (priv->spinner)
-    {
-      gtk_widget_destroy (GTK_WIDGET (priv->spinner));
-      priv->spinner = NULL;
-    }
-
-  if (priv->builder)
-    gtk_widget_queue_draw (WID ("preview-area"));
-
-  /* remove the reference taken when the copy was set up */
-  g_object_unref (panel);
-}
-
 static void
 select_style (GtkComboBox *box,
 	      GDesktopBackgroundStyle new_style)
@@ -363,67 +315,196 @@ select_style (GtkComboBox *box,
 
 static void
 update_preview (CcBackgroundPanelPrivate *priv,
-                GnomeWPItem              *item,
-                gboolean                  redraw_preview)
+                CcBackgroundItem         *item)
 {
   gchar *markup;
   gboolean changes_with_time;
 
   if (item && priv->current_background)
     {
-      if (priv->current_background->pcolor)
-        gdk_color_free (priv->current_background->pcolor);
-      priv->current_background->pcolor = gdk_color_copy (item->pcolor);
-
-      if (priv->current_background->scolor)
-        gdk_color_free (priv->current_background->scolor);
-      priv->current_background->scolor = gdk_color_copy (item->scolor);
-
-      g_free (priv->current_background->filename);
-      priv->current_background->filename = g_strdup (item->filename);
-
-      g_free (priv->current_background->name);
-      priv->current_background->name = g_strdup (item->name);
-
-      priv->current_background->options = item->options;
-      priv->current_background->shade_type = item->shade_type;
-
-      priv->current_background->height = item->height;
-      priv->current_background->width = item->width;
-
-      gnome_wp_item_ensure_gnome_bg (priv->current_background);
-      gnome_wp_item_update_size (priv->current_background, priv->thumb_factory);
+      g_object_unref (priv->current_background);
+      priv->current_background = cc_background_item_copy (item);
+      cc_background_item_load (priv->current_background, NULL);
     }
 
-  source_update_edit_box (priv);
+  source_update_edit_box (priv, FALSE);
 
   changes_with_time = FALSE;
 
   if (priv->current_background)
     {
-      markup = g_strdup_printf ("<b>%s</b>", priv->current_background->name);
+      GdkColor pcolor, scolor;
+
+      markup = g_strdup_printf ("<b>%s</b>", cc_background_item_get_name (priv->current_background));
       gtk_label_set_markup (GTK_LABEL (WID ("background-label")), markup);
       g_free (markup);
 
-      gtk_label_set_text (GTK_LABEL (WID ("size_label")), priv->current_background->size);
+      gtk_label_set_text (GTK_LABEL (WID ("size_label")), cc_background_item_get_size (priv->current_background));
 
-      gtk_color_button_set_color (GTK_COLOR_BUTTON (WID ("style-pcolor")),
-                                  priv->current_background->pcolor);
-      gtk_color_button_set_color (GTK_COLOR_BUTTON (WID ("style-scolor")),
-                                  priv->current_background->scolor);
+      gdk_color_parse (cc_background_item_get_pcolor (priv->current_background), &pcolor);
+      gdk_color_parse (cc_background_item_get_scolor (priv->current_background), &scolor);
+
+      gtk_color_button_set_color (GTK_COLOR_BUTTON (WID ("style-pcolor")), &pcolor);
+      gtk_color_button_set_color (GTK_COLOR_BUTTON (WID ("style-scolor")), &scolor);
 
       select_style (GTK_COMBO_BOX (WID ("style-combobox")),
-                    priv->current_background->options);
+                    cc_background_item_get_placement (priv->current_background));
 
-      if (priv->current_background->bg)
-        changes_with_time = gnome_bg_changes_with_time (priv->current_background->bg);
+      changes_with_time = cc_background_item_changes_with_time (priv->current_background);
     }
 
   gtk_widget_set_visible (WID ("slide_image"), changes_with_time);
   gtk_widget_set_visible (WID ("slide-label"), changes_with_time);
 
-  if (redraw_preview)
-    gtk_widget_queue_draw (WID ("preview-area"));
+  gtk_widget_queue_draw (WID ("preview-area"));
+}
+
+static char *
+get_save_path (void)
+{
+  return g_build_filename (g_get_user_config_dir (),
+			   "gnome-control-center",
+			   "backgrounds",
+			   "last-edited.xml",
+			   NULL);
+}
+
+static gboolean
+create_save_dir (void)
+{
+  char *path;
+
+  path = g_build_filename (g_get_user_config_dir (),
+			   "gnome-control-center",
+			   "backgrounds",
+			   NULL);
+  if (g_mkdir_with_parents (path, 0755) < 0)
+    {
+      g_warning ("Failed to create directory '%s'", path);
+      g_free (path);
+      return FALSE;
+    }
+  g_free (path);
+  return TRUE;
+}
+
+static void
+copy_finished_cb (GObject      *source_object,
+                  GAsyncResult *result,
+                  gpointer      pointer)
+{
+  GError *err = NULL;
+  CcBackgroundPanel *panel = (CcBackgroundPanel *) pointer;
+  CcBackgroundPanelPrivate *priv = panel->priv;
+  CcBackgroundItem *item;
+
+  if (!g_file_copy_finish (G_FILE (source_object), result, &err))
+    {
+      if (err->code != G_IO_ERROR_CANCELLED)
+        g_warning ("Failed to copy image to cache location: %s", err->message);
+
+      g_error_free (err);
+    }
+  item = g_object_get_data (source_object, "item");
+
+  /* the panel may have been destroyed before the callback is run, so be sure
+   * to check the widgets are not NULL */
+
+  if (priv->spinner)
+    {
+      gtk_widget_destroy (GTK_WIDGET (priv->spinner));
+      priv->spinner = NULL;
+    }
+
+  if (priv->current_background)
+    cc_background_item_load (priv->current_background, NULL);
+
+  if (priv->builder)
+    {
+      char *filename;
+
+      update_preview (priv, item);
+
+      /* Save the source XML if there is one */
+      filename = get_save_path ();
+      if (create_save_dir ())
+        cc_background_xml_save (priv->current_background, filename);
+    }
+
+  /* remove the reference taken when the copy was set up */
+  g_object_unref (panel);
+}
+
+static void
+update_remove_button (CcBackgroundPanel *panel,
+		      CcBackgroundItem  *item)
+{
+  CcBackgroundPanelPrivate *priv;
+  const char *uri;
+  char *cache_path;
+  GFile *bg, *cache, *parent;
+  gboolean sensitive = FALSE;
+
+  priv = panel->priv;
+
+  if (priv->current_source != SOURCE_PICTURES)
+    goto bail;
+
+  uri = cc_background_item_get_uri (item);
+  if (uri == NULL)
+    goto bail;
+
+  bg = g_file_new_for_uri (uri);
+  parent = g_file_get_parent (bg);
+  if (parent == NULL)
+    {
+      g_object_unref (bg);
+      goto bail;
+    }
+  cache_path = bg_pictures_source_get_cache_path ();
+  cache = g_file_new_for_path (cache_path);
+  g_free (cache_path);
+
+  if (g_file_equal (parent, cache))
+    sensitive = TRUE;
+
+  g_object_unref (parent);
+  g_object_unref (cache);
+
+bail:
+  gtk_widget_set_sensitive (WID ("remove_button"), sensitive);
+
+}
+
+static CcBackgroundItem *
+get_selected_item (CcBackgroundPanel *panel)
+{
+  CcBackgroundPanelPrivate *priv = panel->priv;
+  GtkIconView *icon_view;
+  GtkTreeIter iter;
+  GtkTreeModel *model;
+  GList *list;
+  CcBackgroundItem *item;
+
+  icon_view = GTK_ICON_VIEW (WID ("backgrounds-iconview"));
+  item = NULL;
+  list = gtk_icon_view_get_selected_items (icon_view);
+
+  if (!list)
+    return NULL;
+
+  model = gtk_icon_view_get_model (icon_view);
+
+  if (gtk_tree_model_get_iter (model, &iter, (GtkTreePath*) list->data) == FALSE)
+    goto bail;
+
+  gtk_tree_model_get (model, &iter, 1, &item, -1);
+
+bail:
+  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
+  g_list_free (list);
+
+  return item;
 }
 
 static void
@@ -431,65 +512,71 @@ backgrounds_changed_cb (GtkIconView       *icon_view,
                         CcBackgroundPanel *panel)
 {
   GtkTreeIter iter;
-  GList *list;
   GtkTreeModel *model;
-  GnomeWPItem *item;
-  gchar *pcolor, *scolor;
+  CcBackgroundItem *item;
   CcBackgroundPanelPrivate *priv = panel->priv;
+  char *pcolor, *scolor;
   gboolean draw_preview = TRUE;
+  const char *uri;
+  CcBackgroundItemFlags flags;
+  char *filename;
 
-  list = gtk_icon_view_get_selected_items (icon_view);
+  item = get_selected_item (panel);
 
-  if (!list)
+  if (item == NULL)
     return;
 
-  /* check if the current source is read only, i.e. the image placement and
-   * color is predefined */
+  /* Update current source */
   model = gtk_combo_box_get_model (GTK_COMBO_BOX (WID ("sources-combobox")));
   gtk_combo_box_get_active_iter (GTK_COMBO_BOX (WID ("sources-combobox")),
                                  &iter);
   gtk_tree_model_get (model, &iter,
-  		      COL_SOURCE_READONLY, &priv->current_source_readonly,
-  		      COL_SOURCE_TYPE, &priv->current_source, -1);
+		      COL_SOURCE_TYPE, &priv->current_source, -1);
 
-  model = gtk_icon_view_get_model (icon_view);
+  uri = cc_background_item_get_uri (item);
+  flags = cc_background_item_get_flags (item);
 
-  gtk_tree_model_get_iter (model, &iter, (GtkTreePath*) list->data);
-
-  g_list_foreach (list, (GFunc)gtk_tree_path_free, NULL);
-  g_list_free (list);
-
-  gtk_tree_model_get (model, &iter, 1, &item, -1);
-
-  if (!g_strcmp0 (item->filename, "(none)"))
+  if ((flags & CC_BACKGROUND_ITEM_HAS_URI) && uri == NULL)
     {
       g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, G_DESKTOP_BACKGROUND_STYLE_NONE);
-      g_settings_set_string (priv->settings, WP_FILE_KEY, "");
+      g_settings_set_string (priv->settings, WP_URI_KEY, "");
     }
-  else if (item->source_url)
+  else if (cc_background_item_get_source_url (item) != NULL &&
+	   cc_background_item_get_needs_download (item))
     {
       GFile *source, *dest;
-      gchar *cache_path;
+      gchar *cache_path, *basename, *dest_path, *display_name, *dest_uri;
       GdkPixbuf *pixbuf;
 
-      cache_path = g_build_filename (g_get_user_cache_dir (),
-                                     "gnome-background",
-                                     NULL);
+      cache_path = bg_pictures_source_get_cache_path ();
+      if (g_mkdir_with_parents (cache_path, 0755) < 0)
+        {
+          g_warning ("Failed to create directory '%s'", cache_path);
+          g_free (cache_path);
+          return;
+	}
+      g_free (cache_path);
 
-      source = g_file_new_for_uri (item->source_url);
-      dest = g_file_new_for_path (cache_path);
+      dest_path = bg_pictures_source_get_unique_path (cc_background_item_get_source_url (item));
+      dest = g_file_new_for_path (dest_path);
+      g_free (dest_path);
+      source = g_file_new_for_uri (cc_background_item_get_source_url (item));
+      basename = g_file_get_basename (source);
+      display_name = g_filename_display_name (basename);
+      dest_path = g_file_get_path (dest);
+      g_free (basename);
 
       /* create a blank image to use until the source image is ready */
       pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 1, 1);
       gdk_pixbuf_fill (pixbuf, 0x00000000);
-      gdk_pixbuf_save (pixbuf, cache_path, "png", NULL, NULL);
+      gdk_pixbuf_save (pixbuf, dest_path, "png", NULL, NULL);
       g_object_unref (pixbuf);
+      g_free (dest_path);
 
       if (priv->copy_cancellable)
         {
           g_cancellable_cancel (priv->copy_cancellable);
           g_cancellable_reset (priv->copy_cancellable);
-
         }
 
       if (priv->spinner)
@@ -508,69 +595,89 @@ backgrounds_changed_cb (GtkIconView       *icon_view,
       /* reference the panel in case it is removed before the copy is
        * finished */
       g_object_ref (panel);
+      g_object_set_data_full (G_OBJECT (source), "item", g_object_ref (item), g_object_unref);
       g_file_copy_async (source, dest, G_FILE_COPY_OVERWRITE,
                          G_PRIORITY_DEFAULT, priv->copy_cancellable,
                          NULL, NULL,
                          copy_finished_cb, panel);
+      g_object_unref (source);
+      dest_uri = g_file_get_uri (dest);
+      g_object_unref (dest);
 
-      g_settings_set_string (priv->settings, WP_FILE_KEY, cache_path);
-      g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, item->options);
-      g_free (item->filename);
-      item->filename = cache_path;
+      g_settings_set_string (priv->settings, WP_URI_KEY, dest_uri);
+      g_object_set (G_OBJECT (item),
+		    "uri", dest_uri,
+		    "needs-download", FALSE,
+		    "name", display_name,
+		    NULL);
+      g_free (display_name);
+      g_free (dest_uri);
 
       /* delay the updated drawing of the preview until the copy finishes */
       draw_preview = FALSE;
     }
   else
     {
-       gchar *uri;
-
-       if (g_utf8_validate (item->filename, -1, NULL))
-         uri = g_strdup (item->filename);
-       else
-         uri = g_filename_to_utf8 (item->filename, -1, NULL, NULL, NULL);
-
-       if (uri == NULL)
-         {
-           g_warning ("Failed to convert filename to UTF-8: %s",
-                      item->filename);
-         }
-       else
-         {
-	   g_settings_set_string (priv->settings, WP_FILE_KEY, uri);
-           g_free (uri);
-         }
-
-       g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, item->options);
+      g_settings_set_string (priv->settings, WP_URI_KEY, uri);
     }
 
-  g_settings_set_enum (priv->settings, WP_SHADING_KEY, item->shade_type);
-
-  /* When changing for another colour, don't overwrite what's
-   * in GSettings, but read from it instead */
-  if (priv->current_source == SOURCE_COLORS)
+  /* Also set the placement if we have a URI and the previous value was none */
+  if (flags & CC_BACKGROUND_ITEM_HAS_PLACEMENT)
     {
-      pcolor = g_settings_get_string (priv->settings, WP_PCOLOR_KEY);
-      scolor = g_settings_get_string (priv->settings, WP_SCOLOR_KEY);
-      gdk_color_parse (pcolor, item->pcolor);
-      gdk_color_parse (scolor, item->scolor);
+      g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, cc_background_item_get_placement (item));
+    }
+  else if (uri != NULL)
+    {
+      GDesktopBackgroundStyle style;
+      style = g_settings_get_enum (priv->settings, WP_OPTIONS_KEY);
+      if (style == G_DESKTOP_BACKGROUND_STYLE_NONE)
+        g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, cc_background_item_get_placement (item));
+    }
+
+  if (flags & CC_BACKGROUND_ITEM_HAS_SHADING)
+    g_settings_set_enum (priv->settings, WP_SHADING_KEY, cc_background_item_get_shading (item));
+
+  /* When changing to a background with colours set,
+   * don't overwrite what's in GSettings, but read
+   * from it instead.
+   * We have a hack for the colors source though */
+  if (flags & CC_BACKGROUND_ITEM_HAS_PCOLOR &&
+      priv->current_source != SOURCE_COLORS)
+    {
+      g_settings_set_string (priv->settings, WP_PCOLOR_KEY, cc_background_item_get_pcolor (item));
     }
   else
     {
-      pcolor = gdk_color_to_string (item->pcolor);
-      scolor = gdk_color_to_string (item->scolor);
-      g_settings_set_string (priv->settings, WP_PCOLOR_KEY, pcolor);
-      g_settings_set_string (priv->settings, WP_SCOLOR_KEY, scolor);
+      pcolor = g_settings_get_string (priv->settings, WP_PCOLOR_KEY);
+      g_object_set (G_OBJECT (item), "primary-color", pcolor, NULL);
     }
 
-  g_free (pcolor);
-  g_free (scolor);
+  if (flags & CC_BACKGROUND_ITEM_HAS_SCOLOR &&
+      priv->current_source != SOURCE_COLORS)
+    {
+      g_settings_set_string (priv->settings, WP_SCOLOR_KEY, cc_background_item_get_scolor (item));
+    }
+  else
+    {
+      scolor = g_settings_get_string (priv->settings, WP_SCOLOR_KEY);
+      g_object_set (G_OBJECT (item), "secondary-color", scolor, NULL);
+    }
 
   /* Apply all changes */
   g_settings_apply (priv->settings);
 
+  update_remove_button (panel, item);
+
   /* update the preview information */
-  update_preview (priv, item, draw_preview);
+  if (draw_preview != FALSE)
+    {
+      update_preview (priv, item);
+
+      /* Save the source XML if there is one */
+      filename = get_save_path ();
+      if (create_save_dir ())
+        cc_background_xml_save (priv->current_background, filename);
+    }
 }
 
 static gboolean
@@ -593,7 +700,7 @@ preview_draw_cb (GtkWidget         *widget,
   if (priv->current_background)
     {
       GIcon *icon;
-      icon = gnome_wp_item_get_frame_thumbnail (priv->current_background,
+      icon = cc_background_item_get_frame_thumbnail (priv->current_background,
                                                 priv->thumb_factory,
                                                 preview_width,
                                                 preview_height,
@@ -668,11 +775,11 @@ style_changed_cb (GtkComboBox       *box,
   g_settings_set_enum (priv->settings, WP_OPTIONS_KEY, value);
 
   if (priv->current_background)
-    priv->current_background->options = value;
+    g_object_set (G_OBJECT (priv->current_background), "placement", value, NULL);
 
   g_settings_apply (priv->settings);
 
-  update_preview (priv, NULL, TRUE);
+  update_preview (priv, NULL);
 }
 
 static void
@@ -688,38 +795,297 @@ color_changed_cb (GtkColorButton    *button,
   if (WID ("style-pcolor") == GTK_WIDGET (button))
     is_pcolor = TRUE;
 
-  if (priv->current_background)
-    {
-      if (is_pcolor)
-        *priv->current_background->pcolor = color;
-      else
-        *priv->current_background->scolor = color;
-    }
-
   value = gdk_color_to_string (&color);
 
-  if (is_pcolor)
-    g_settings_set_string (priv->settings, WP_PCOLOR_KEY, value);
-  else
-    g_settings_set_string (priv->settings, WP_SCOLOR_KEY, value);
+  if (priv->current_background)
+    {
+      g_object_set (G_OBJECT (priv->current_background),
+		    is_pcolor ? "primary-color" : "secondary-color", value, NULL);
+    }
+
+  g_settings_set_string (priv->settings,
+			 is_pcolor ? WP_PCOLOR_KEY : WP_SCOLOR_KEY, value);
 
   g_settings_apply (priv->settings);
 
   g_free (value);
 
-  update_preview (priv, NULL, TRUE);
+  update_preview (priv, NULL);
+}
+
+static void
+row_inserted (GtkTreeModel      *tree_model,
+	      GtkTreePath       *path,
+	      GtkTreeIter       *iter,
+	      CcBackgroundPanel *panel)
+{
+  GtkListStore *store;
+  CcBackgroundPanelPrivate *priv;
+
+  priv = panel->priv;
+
+  store = bg_source_get_liststore (BG_SOURCE (panel->priv->pictures_source));
+  g_signal_handlers_disconnect_by_func (G_OBJECT (store), G_CALLBACK (row_inserted), panel);
+
+  /* Change source */
+  gtk_combo_box_set_active (GTK_COMBO_BOX (WID ("sources-combobox")), SOURCE_PICTURES);
+
+  /* And select the newly added item */
+  gtk_icon_view_select_path (GTK_ICON_VIEW (WID ("backgrounds-iconview")), path);
+}
+
+static void
+add_custom_wallpaper (CcBackgroundPanel *panel,
+		      const char        *uri)
+{
+  GtkListStore *store;
+
+  store = bg_source_get_liststore (BG_SOURCE (panel->priv->pictures_source));
+  g_signal_connect (G_OBJECT (store), "row-inserted",
+		    G_CALLBACK (row_inserted), panel);
+
+  if (bg_pictures_source_add (panel->priv->pictures_source, uri) == FALSE) {
+    g_signal_handlers_disconnect_by_func (G_OBJECT (store), G_CALLBACK (row_inserted), panel);
+    return;
+  }
+
+  /* Wait for the item to get added */
+}
+
+static void
+file_chooser_response (GtkDialog         *chooser,
+                       gint               response,
+                       CcBackgroundPanel *panel)
+{
+  char *uri;
+
+  if (response != GTK_RESPONSE_ACCEPT)
+    {
+      gtk_widget_destroy (GTK_WIDGET (chooser));
+      return;
+    }
+
+  uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (chooser));
+  gtk_widget_destroy (GTK_WIDGET (chooser));
+
+  add_custom_wallpaper (panel, uri);
+  g_free (uri);
+}
+
+static void
+update_chooser_preview (GtkFileChooser    *chooser,
+			CcBackgroundPanel *panel)
+{
+  GnomeDesktopThumbnailFactory *thumb_factory;
+  char *uri;
+
+  thumb_factory = panel->priv->thumb_factory;
+
+  uri = gtk_file_chooser_get_preview_uri (chooser);
+
+  if (uri)
+    {
+      GdkPixbuf *pixbuf = NULL;
+      const gchar *mime_type = NULL;
+      GFile *file;
+      GFileInfo *file_info;
+      GtkWidget *preview;
+
+      preview = gtk_file_chooser_get_preview_widget (chooser);
+
+      file = g_file_new_for_uri (uri);
+      file_info = g_file_query_info (file,
+				     G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
+				     G_FILE_QUERY_INFO_NONE,
+				     NULL, NULL);
+      g_object_unref (file);
+
+      if (file_info != NULL) {
+	      mime_type = g_file_info_get_content_type (file_info);
+	      g_object_unref (file_info);
+      }
+
+      if (mime_type)
+        {
+        pixbuf = gnome_desktop_thumbnail_factory_generate_thumbnail (thumb_factory,
+								     uri,
+								     mime_type);
+	}
+
+      gtk_dialog_set_response_sensitive (GTK_DIALOG (chooser),
+					 GTK_RESPONSE_ACCEPT,
+					 (pixbuf != NULL));
+
+      if (pixbuf != NULL)
+        {
+          gtk_image_set_from_pixbuf (GTK_IMAGE (preview), pixbuf);
+	  g_object_unref (pixbuf);
+	}
+      else
+        {
+          gtk_image_set_from_stock (GTK_IMAGE (preview),
+				    GTK_STOCK_DIALOG_QUESTION,
+				    GTK_ICON_SIZE_DIALOG);
+	}
+
+      if (bg_pictures_source_is_known (panel->priv->pictures_source, uri))
+        gtk_dialog_set_response_sensitive (GTK_DIALOG (chooser), GTK_RESPONSE_ACCEPT, FALSE);
+      else
+        gtk_dialog_set_response_sensitive (GTK_DIALOG (chooser), GTK_RESPONSE_ACCEPT, TRUE);
+
+      g_free (uri);
+    }
+
+  gtk_file_chooser_set_preview_widget_active (chooser, TRUE);
+}
+
+static void
+add_button_clicked (GtkButton         *button,
+		    CcBackgroundPanel *panel)
+{
+  GtkWidget *chooser;
+  const gchar *folder;
+  GtkWidget *preview;
+  GtkFileFilter *filter;
+  CcBackgroundPanelPrivate *priv;
+
+  priv = panel->priv;
+
+  filter = gtk_file_filter_new ();
+  gtk_file_filter_add_mime_type (filter, "image/png");
+  gtk_file_filter_add_mime_type (filter, "image/jpeg");
+
+  chooser = gtk_file_chooser_dialog_new (_("Browse for more pictures"),
+					 GTK_WINDOW (gtk_widget_get_toplevel (WID ("background-panel"))),
+					 GTK_FILE_CHOOSER_ACTION_OPEN,
+					 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+					 GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+					 NULL);
+  gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (chooser), filter);
+
+  gtk_window_set_modal (GTK_WINDOW (chooser), TRUE);
+
+  preview = gtk_image_new ();
+  gtk_widget_set_size_request (preview, 128, -1);
+  gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (chooser), preview);
+  gtk_file_chooser_set_use_preview_label (GTK_FILE_CHOOSER (chooser), FALSE);
+  gtk_widget_show (preview);
+  g_signal_connect (chooser, "update-preview",
+		    G_CALLBACK (update_chooser_preview), panel);
+
+  folder = g_get_user_special_dir (G_USER_DIRECTORY_PICTURES);
+  if (folder)
+    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (chooser),
+					 folder);
+
+  g_signal_connect (chooser, "response",
+		    G_CALLBACK (file_chooser_response), panel);
+
+  gtk_window_present (GTK_WINDOW (chooser));
+}
+
+static void
+remove_button_clicked (GtkButton         *button,
+		       CcBackgroundPanel *panel)
+{
+  CcBackgroundItem *item;
+  GtkListStore *store;
+  GtkTreePath *path;
+  CcBackgroundPanelPrivate *priv;
+
+  priv = panel->priv;
+
+  item = get_selected_item (panel);
+  if (item == NULL)
+    g_assert_not_reached ();
+
+  bg_pictures_source_remove (panel->priv->pictures_source, item);
+  g_object_unref (item);
+
+  /* Are there any items left in the pictures tree store? */
+  store = bg_source_get_liststore (BG_SOURCE (panel->priv->pictures_source));
+  if (gtk_tree_model_iter_n_children (GTK_TREE_MODEL (store), NULL) == 0)
+    gtk_combo_box_set_active (GTK_COMBO_BOX (WID ("sources-combobox")), SOURCE_WALLPAPERS);
+
+  path = gtk_tree_path_new_from_string ("0");
+  gtk_icon_view_select_path (GTK_ICON_VIEW (WID ("backgrounds-iconview")), path);
+  gtk_tree_path_free (path);
+}
+
+static void
+load_current_bg (CcBackgroundPanel *self)
+{
+  CcBackgroundPanelPrivate *priv;
+  CcBackgroundItem *saved, *configured;
+  gchar *uri, *pcolor, *scolor;
+
+  priv = self->priv;
+
+  /* Load the saved configuration */
+  uri = get_save_path ();
+  saved = cc_background_xml_get_item (uri);
+  g_free (uri);
+
+  /* initalise the current background information from settings */
+  uri = g_settings_get_string (priv->settings, WP_URI_KEY);
+  if (uri && *uri == '\0')
+    {
+      g_free (uri);
+      uri = NULL;
+    }
+  else
+    {
+      GFile *file;
+
+      file = g_file_new_for_commandline_arg (uri);
+      g_object_unref (file);
+    }
+  configured = cc_background_item_new (uri);
+  g_free (uri);
+
+  pcolor = g_settings_get_string (priv->settings, WP_PCOLOR_KEY);
+  scolor = g_settings_get_string (priv->settings, WP_SCOLOR_KEY);
+  g_object_set (G_OBJECT (configured),
+		"name", _("Current background"),
+		"placement", g_settings_get_enum (priv->settings, WP_OPTIONS_KEY),
+		"shading", g_settings_get_enum (priv->settings, WP_SHADING_KEY),
+		"primary-color", pcolor,
+		"secondary-color", scolor,
+		NULL);
+  g_free (pcolor);
+  g_free (scolor);
+
+  if (saved != NULL && cc_background_item_compare (saved, configured))
+    {
+      CcBackgroundItemFlags flags;
+      flags = cc_background_item_get_flags (saved);
+      /* Special case for colours */
+      if (cc_background_item_get_placement (saved) == G_DESKTOP_BACKGROUND_STYLE_NONE)
+        flags &=~ (CC_BACKGROUND_ITEM_HAS_PCOLOR | CC_BACKGROUND_ITEM_HAS_SCOLOR);
+      g_object_set (G_OBJECT (configured),
+		    "name", cc_background_item_get_name (saved),
+		    "flags", flags,
+		    "source-url", cc_background_item_get_source_url (saved),
+		    "source-xml", cc_background_item_get_source_xml (saved),
+		    NULL);
+    }
+  if (saved != NULL)
+    g_object_unref (saved);
+
+  priv->current_background = configured;
+  cc_background_item_load (priv->current_background, NULL);
 }
 
 static void
 cc_background_panel_init (CcBackgroundPanel *self)
 {
   CcBackgroundPanelPrivate *priv;
-  gchar *objects[] = { "backgrounds-liststore", "style-liststore",
+  gchar *objects[] = { "style-liststore",
       "sources-liststore", "background-panel", "sizegroup", NULL };
   GError *err = NULL;
   GtkWidget *widget;
   GtkListStore *store;
-  gchar *filename;
+  GtkStyleContext *context;
 
   priv = self->priv = BACKGROUND_PANEL_PRIVATE (self);
 
@@ -746,7 +1112,6 @@ cc_background_panel_init (CcBackgroundPanel *self)
   gtk_list_store_insert_with_values (store, NULL, G_MAXINT,
                                      COL_SOURCE_NAME, _("Wallpapers"),
                                      COL_SOURCE_TYPE, SOURCE_WALLPAPERS,
-                                     COL_SOURCE_READONLY, TRUE,
                                      COL_SOURCE, priv->wallpapers_source,
                                      -1);
 
@@ -754,7 +1119,6 @@ cc_background_panel_init (CcBackgroundPanel *self)
   gtk_list_store_insert_with_values (store, NULL, G_MAXINT,
                                      COL_SOURCE_NAME, _("Pictures Folder"),
                                      COL_SOURCE_TYPE, SOURCE_PICTURES,
-                                     COL_SOURCE_READONLY, FALSE,
                                      COL_SOURCE, priv->pictures_source,
                                      -1);
 
@@ -762,7 +1126,6 @@ cc_background_panel_init (CcBackgroundPanel *self)
   gtk_list_store_insert_with_values (store, NULL, G_MAXINT,
                                      COL_SOURCE_NAME, _("Colors & Gradients"),
                                      COL_SOURCE_TYPE, SOURCE_COLORS,
-                                     COL_SOURCE_READONLY, FALSE,
                                      COL_SOURCE, priv->colors_source,
                                      -1);
 
@@ -771,15 +1134,13 @@ cc_background_panel_init (CcBackgroundPanel *self)
   gtk_list_store_insert_with_values (store, NULL, G_MAXINT,
                                      COL_SOURCE_NAME, _("Flickr"),
                                      COL_SOURCE_TYPE, SOURCE_FLICKR,
-                                     COL_SOURCE_READONLY, FALSE,
                                      COL_SOURCE, priv->flickr_source,
                                      -1);
 #endif
 
 
   /* add the top level widget */
-  widget = (GtkWidget*)
-    gtk_builder_get_object (priv->builder, "background-panel");
+  widget = WID ("background-panel");
 
   gtk_container_add (GTK_CONTAINER (self), widget);
   gtk_widget_show_all (GTK_WIDGET (self));
@@ -797,7 +1158,21 @@ cc_background_panel_init (CcBackgroundPanel *self)
                     G_CALLBACK (backgrounds_changed_cb),
                     self);
 
+  /* Join treeview and buttons */
+  widget = WID ("scrolledwindow1");
+  context = gtk_widget_get_style_context (widget);
+  gtk_style_context_set_junction_sides (context, GTK_JUNCTION_BOTTOM);
+  widget = WID ("toolbar1");
+  context = gtk_widget_get_style_context (widget);
+  gtk_style_context_set_junction_sides (context, GTK_JUNCTION_TOP);
+
+  g_signal_connect (WID ("add_button"), "clicked",
+		    G_CALLBACK (add_button_clicked), self);
+  g_signal_connect (WID ("remove_button"), "clicked",
+		    G_CALLBACK (remove_button_clicked), self);
+
   /* setup preview area */
+  gtk_label_set_ellipsize (GTK_LABEL (WID ("background-label")), PANGO_ELLIPSIZE_END);
   widget = WID ("preview-area");
   g_signal_connect (widget, "draw", G_CALLBACK (preview_draw_cb),
                     self);
@@ -820,26 +1195,12 @@ cc_background_panel_init (CcBackgroundPanel *self)
 
   priv->thumb_factory = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
 
-  /* initalise the current background information from settings */
-  filename = g_settings_get_string (priv->settings, WP_FILE_KEY);
-  if (!filename || !g_strcmp0 (filename, ""))
-    {
-      g_free (filename);
-      filename = g_strdup ("(none)");
-    }
+  load_current_bg (self);
 
-  priv->current_background = g_new0 (GnomeWPItem, 1);
-  priv->current_background->filename = filename;
-  priv->current_background->name = g_strdup (_("Current background"));
-
-  gnome_wp_item_update (priv->current_background);
-  gnome_wp_item_ensure_gnome_bg (priv->current_background);
-  gnome_wp_item_update_size (priv->current_background, priv->thumb_factory);
-
-  update_preview (priv, NULL, TRUE);
+  update_preview (priv, NULL);
 
   /* Setup the edit box with our current settings */
-  setup_edit_box (priv);
+  source_update_edit_box (priv, TRUE);
 }
 
 void
