@@ -19,6 +19,8 @@
  *
  */
 
+#include <config.h>
+
 #include "cc-info-panel.h"
 
 #include <sys/vfs.h>
@@ -43,6 +45,14 @@ G_DEFINE_DYNAMIC_TYPE (CcInfoPanel, cc_info_panel, CC_TYPE_PANEL)
 #define INFO_PANEL_PRIVATE(o) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), CC_TYPE_INFO_PANEL, CcInfoPanelPrivate))
 
+typedef struct {
+  /* Will be one of the other two below, or "Unknown" */ 
+  const char *hardware_string;
+
+  char *xorg_vesa_hardware;
+  char *glx_renderer;
+} GraphicsData;
+
 struct _CcInfoPanelPrivate
 {
   GtkBuilder    *builder;
@@ -56,6 +66,8 @@ struct _CcInfoPanelPrivate
   GDBusProxy    *pk_proxy;
   GDBusProxy    *pk_transaction_proxy;
   GSettings     *session_settings;
+
+  GraphicsData  *graphics_data;
 };
 
 typedef struct
@@ -233,61 +245,16 @@ prettify_info (const char *info)
   return pretty;
 }
 
-
-static char *
-get_graphics_info_lspci (void)
+static void
+graphics_data_free (GraphicsData *gdata)
 {
-  GError     *error;
-  GRegex     *re;
-  GMatchInfo *match_info;
-  char       *output;
-  char       *result;
-  GString    *info;
-
-  info = g_string_new (NULL);
-
-  error = NULL;
-  g_spawn_command_line_sync ("lspci -nn", &output, NULL, NULL, &error);
-  if (error != NULL)
-    {
-      g_warning ("Unable to get graphics info: %s", error->message);
-      g_error_free (error);
-      return NULL;
-    }
-
-  re = g_regex_new ("^[^ ]+ VGA compatible controller [^:]*: ([^([]+).*$", G_REGEX_MULTILINE, 0, &error);
-  if (re == NULL)
-    {
-      g_warning ("Error building regex: %s", error->message);
-      g_error_free (error);
-      goto out;
-    }
-
-  g_regex_match (re, output, 0, &match_info);
-  while (g_match_info_matches (match_info))
-    {
-      char *device;
-
-      device = g_match_info_fetch (match_info, 1);
-      g_string_append_printf (info, "%s ", device);
-      g_free (device);
-
-      g_match_info_next (match_info, NULL);
-    }
-
-  g_match_info_free (match_info);
-  g_regex_unref (re);
-
- out:
-  g_free (output);
-  result = prettify_info (info->str);
-  g_string_free (info, TRUE);
-
-  return result;
+  g_free (gdata->xorg_vesa_hardware);
+  g_free (gdata->glx_renderer);
+  g_slice_free (GraphicsData, gdata);
 }
 
 static char *
-get_graphics_info_glxinfo (void)
+get_graphics_data_glx_renderer (void)
 {
   GError     *error;
   GRegex     *re;
@@ -338,15 +305,91 @@ get_graphics_info_glxinfo (void)
 }
 
 static char *
-get_graphics_info (void)
+get_graphics_data_xorg_vesa_hardware (void)
 {
-  gchar *info;
+  char *display_num;
+  char *log_path;
+  char *log_contents;
+  gsize log_len;
+  GError *error = NULL;
+  GRegex *re;
+  GMatchInfo *match;
+  char *result = NULL;
 
-  info = get_graphics_info_glxinfo ();
-  if (info == NULL)
-    info = get_graphics_info_lspci ();
+  {
+    const char *display;
 
-  return info;
+    display = g_getenv ("DISPLAY");
+    if (!display)
+      return NULL;
+
+    re = g_regex_new ("^:([0-9]+)", 0, 0, NULL);
+    g_assert (re != NULL);
+
+    g_regex_match (re, display, 0, &match);
+
+    if (!g_match_info_matches (match))
+      {
+        g_regex_unref (re);
+        g_match_info_free (match);
+        return NULL;
+      }
+
+    display_num = g_match_info_fetch (match, 1);
+
+    g_regex_unref (re);
+    re = NULL;
+    g_match_info_free (match);
+    match = NULL;
+  }
+
+  log_path = g_strdup_printf ("/var/log/Xorg.%s.log", display_num);
+  g_free (display_num);
+  log_contents = NULL;
+  g_file_get_contents (log_path, &log_contents, &log_len, &error);
+  g_free (log_path);
+  if (!log_contents)
+    return NULL;
+
+  re = g_regex_new ("VESA VBE OEM Product: (.*)$", G_REGEX_MULTILINE, 0, NULL);
+  g_assert (re != NULL);
+
+  g_regex_match (re, log_contents, 0, &match);
+  if (g_match_info_matches (match))
+    {
+      char *tmp;
+      char *pretty_tmp;
+      tmp = g_match_info_fetch (match, 1);
+      pretty_tmp = prettify_info (tmp);
+      g_free (tmp);
+      /* Translators: VESA is an techncial acronym, don't translate it. */
+      result = g_strdup_printf (_("VESA: %s"), pretty_tmp); 
+      g_free (pretty_tmp);
+    }
+  g_match_info_free (match);
+  g_regex_unref (re);
+
+  return result;
+}
+
+static GraphicsData *
+get_graphics_data (void)
+{
+  GraphicsData *result;
+
+  result = g_slice_new0 (GraphicsData);
+
+  result->glx_renderer = get_graphics_data_glx_renderer ();
+  result->xorg_vesa_hardware = get_graphics_data_xorg_vesa_hardware ();
+
+  if (result->xorg_vesa_hardware != NULL)
+    result->hardware_string = result->xorg_vesa_hardware;
+  else if (result->glx_renderer != NULL)
+    result->hardware_string = result->glx_renderer;
+  else
+    result->hardware_string = _("Unknown");
+
+  return result;
 }
 
 static gboolean
@@ -431,6 +474,12 @@ cc_info_panel_dispose (GObject *object)
       priv->pk_transaction_proxy = NULL;
     }
 
+  if (priv->graphics_data != NULL)
+    {
+      graphics_data_free (priv->graphics_data);
+      priv->graphics_data = NULL;
+    }
+
   G_OBJECT_CLASS (cc_info_panel_parent_class)->dispose (object);
 }
 
@@ -753,15 +802,8 @@ info_panel_setup_graphics (CcInfoPanel  *self)
   GtkSwitch *sw;
   char *text;
 
-  text = get_graphics_info ();
-  widget = WID (self->priv->builder, "graphics_chipset_label");
-  gtk_label_set_markup (GTK_LABEL (widget), text ? text : "");
-  g_free (text);
-
-  text = NULL;
   widget = WID (self->priv->builder, "graphics_driver_label");
-  gtk_label_set_markup (GTK_LABEL (widget), text ? text : "");
-  g_free (text);
+  gtk_label_set_markup (GTK_LABEL (widget), self->priv->graphics_data->hardware_string);
 
   self->priv->is_fallback = get_current_is_fallback (self);
   if (self->priv->is_fallback)
@@ -979,10 +1021,8 @@ info_panel_setup_overview (CcInfoPanel  *self)
   gtk_label_set_text (GTK_LABEL (widget), text ? text : "");
   g_free (text);
 
-  text = get_graphics_info ();
   widget = WID (self->priv->builder, "graphics_label");
-  gtk_label_set_markup (GTK_LABEL (widget), text ? text : "");
-  g_free (text);
+  gtk_label_set_markup (GTK_LABEL (widget), self->priv->graphics_data->hardware_string);
 
   widget = WID (self->priv->builder, "info_vbox");
   gtk_widget_reparent (widget, (GtkWidget *) self);
@@ -1181,6 +1221,8 @@ cc_info_panel_init (CcInfoPanel *self)
       return;
     }
 
+  self->priv->graphics_data = get_graphics_data ();
+
   widget = WID (self->priv->builder, "updates_button");
   g_signal_connect (widget, "clicked", G_CALLBACK (on_updates_button_clicked), self);
 
